@@ -1,6 +1,8 @@
+from copy import deepcopy
+
 import apache_beam as beam
-from apache_beam.transforms.userstate import CombiningValueStateSpec
 from apache_beam.coders import VarIntCoder
+from apache_beam.transforms.userstate import CombiningValueStateSpec
 
 
 class _PairWithIndexNumberDoFn(beam.DoFn):
@@ -138,7 +140,7 @@ def CreateCategoricalDicts(pcoll, cat_cols, existing_dict_rows):
 
 @beam.ptransform_fn
 def ReplaceCategoricalColumns(
-    inputs, cat_cols, categorical_dict_rows, default_unseen=None
+    inputs, cat_cols, categorical_dict_rows, default_unseen=None, features_key=None
 ):
     """
     Utilizes the "categorical dictionary rows" generated by
@@ -157,16 +159,43 @@ def ReplaceCategoricalColumns(
         | "create dict per column" >> beam.MapTuple(lambda k, v: (k, dict(v)))
     )
 
-    def _process(row, d):
-        ret = {}
-        for col, val in row.items():
-            if col in cat_cols:
-                ret[col] = d.get(col, {}).get(val, default_unseen)
-            else:
-                ret[col] = val
-        return ret
+    if features_key:
+        # Nesting -- process separately from row
+        def _process_nested(row, d):
+            digested = {}
 
-    return inputs | beam.Map(
-        _process,
-        d=beam.pvalue.AsDict(categorical_mapping_dicts),
-    )
+            features_items = row[features_key]
+
+            for col, val in features_items.items():
+                if col in cat_cols:
+                    digested[col] = d.get(col, {}).get(val, default_unseen)
+                else:
+                    digested[col] = val
+
+            ret = deepcopy(row)
+            ret[features_key] = digested
+
+            return ret
+
+        return inputs | beam.Map(
+            _process_nested,
+            d=beam.pvalue.AsDict(categorical_mapping_dicts),
+        )
+
+    else:
+        # No nesting - use row in place
+        def _process(row, d):
+            digested = {}
+
+            for col, val in row.items():
+                if col in cat_cols:
+                    digested[col] = d.get(col, {}).get(val, default_unseen)
+                else:
+                    digested[col] = val
+
+            return digested
+
+        return inputs | beam.Map(
+            _process,
+            d=beam.pvalue.AsDict(categorical_mapping_dicts),
+        )
